@@ -1,140 +1,11 @@
 package cyrillic;
-$cyrillic::VERSION = '1.23';
-
-=head1 NAME
-
-cyrillic - Library for fast and easy cyrillic text manipulation
-
-=head1 SYNOPSIS
-
-  use cyrillic qw/866 win2dos convert locase upcase detect/;
-
-  print convert( 866, 1251, $str );
-  print convert( 'dos','win', \$str );
-  print win2dos $str;
-
-=head1 DESCRIPTION
-
-If first import parameter is number of codepage then locale switched to they codepage.
-Specialisation (like 'win2dos') call faster then 'convert'.
-Easy adding new codepage. For they need only add codepage string.
-
-=head1 FUNCTIONS
-
-Library includes converting and helper functions:
-    convert, upcase, locase, upfirst, lofirst;
-    detect, charset.
-
-
-At importing list might be listed named convertors. For Ex.:
-
-  use cyrillic qw/dos2win win2koi mac2dos ibm2dos/;
-
-
-The following rules are correct for converting functions:
-
-  VAR may be SCALAR or REF to SCALAR.
-  If VAR is REF to SCALAR then SCALAR will be converted.
-  If VAR is ommited then $_ operated.
-  If function called to void context and VAR is not REF
-  then result placed to $_.
-
-
-=item B<convert> SRC_CP, DST_CP, [VAR]
-
-Convert VAR from SRC_CP codepage to DST_CP codepage and returns
-converted string.
-
-=item B<upcase> CODEPAGE, [VAR]
-
-Convert VAR to uppercase using CODEPAGE table and returns
-converted string.
-
-=item B<locase> CODEPAGE, [VAR]
-
-Convert VAR to lowercase using CODEPAGE table and returns
-converted string.
-
-=item B<upfirst> CODEPAGE, [VAR]
-
-Convert first char of VAR to uppercase using CODEPAGE table and returns
-converted string.
-
-=item B<lofirst> CODEPAGE, [VAR]
-
-Convert first char of VAR to lowercase using CODEPAGE table and returns
-converted string.
-
-=item B<charset> CODEPAGE
-
-Returns charset name for CODEPAGE.
-
-=item B<detect> ARRAY
-
-Detect codepage of data in ARRAY and returns codepage number.
-If codepage not detected then returns undefined value;
-
-=head1 EXAMPLES
-
-  use cyrillic qw/convert locase upcase detect dos2win win2dos/;
-
-  $\ = "\n";
-  $_ = "\x8F\xE0\xA8\xA2\xA5\xE2 \xF0\xA6\x88\xAA\x88!";
-
-  print; upcase 866;
-  print; dos2win;
-  print; win2dos;
-  print; locase 866;
-  print;
-  print detect $_;
-
-
-  # EQVIVALENT CALLS:
-
-  dos2win( $str );        # called to void context -> result placed to $_
-  $_ = dos2win( $str );
-
-  dos2win( \$str );       # called with REF to string -> direct converting
-  $str = dos2win( $str );
-
-  dos2win();              # with ommited param called -> $_ converted
-  dos2win( \$_ );
-  $_ = dos2win( $_ );
-
-
-  # FOR EASY SWITCH LOCALE CODEPAGE
-
-  use cyrillic qw/866/;   # locale switched to 866 codepage
-
-  use locale;
-  print $str =~ /(\w+)/;
-
-  no locale;
-  print $str =~ /(\w+)/;
-
-=head1 AUTHOR
-
-Albert MICHEEV <Albert@f80.n5049.z2.fidonet.org>
-
-=head1 COPYRIGHT
-
-Copyright (C) 2000, Albert MICHEEV
-
-This module is free software; you can redistribute it or modify it
-under the same terms as Perl itself.
-
-=head1 AVAILABILITY
-
-The latest version of this library is likely to be available from:
-
- http://www.perl.com/CPAN
-
-=cut
+$cyrillic::VERSION = '2.00';
 
 use 5.6.0;
-use vars qw/%CP_NAME %CODEPAGE %STATISTIC $STATISTIC $TRANSLATOR $TRANSLATOR_FIRST/;
+use strict;
+use vars qw/%CP_NAME %CODEPAGE %STATISTIC $STATISTIC $MUTATOR $MUTABLE/;
 
-sub prepare
+sub __prepare
 {
     my($src, $dst)=map $$_[2], @CODEPAGE{@_};
     substr($src, length $1, length $2) = '',
@@ -143,50 +14,95 @@ sub prepare
     return $src, $dst;
 }
 
-sub convert($$;$)
+sub __cs2cs_factory($$)        
 {
-    my ($src, $dst) = (shift, shift);
-    exists $CP_NAME{$_} and $_ = $CP_NAME{$_} for $src, $dst;
-    exists $CODEPAGE{$_} or die"Unknown codepage '$_'\n" for $src, $dst;
-    my $fn = $CODEPAGE{$src}[0].'2'.$CODEPAGE{$dst}[0];
-    *$fn = eval sprintf $TRANSLATOR, prepare($src, $dst) unless defined *$fn;
+    my ($src, $dst, $fn, $sw) = @_;
+    no strict qw/refs/;
+
+    if( $src eq 'utf' &&  $dst eq 'uni'      or $sw =    $src eq 'uni' &&  $dst eq 'utf' )
+    {
+        $fn = $sw ? 'uni2utf' : 'utf2uni';
+        last if defined &$fn;
+
+        eval "use Unicode::String" unless defined $Unicode::String::VERSION;
+
+        *$fn = eval sprintf $MUTATOR, sprintf
+            '%s = Unicode::String::%s( %s )->%s',
+            $MUTABLE, $sw?'utf16':'utf8', $MUTABLE, $sw?'utf8':'utf16';
+    }
+    elsif( $src eq 'utf' ||  $src eq 'uni'    or $sw =    $dst eq 'utf' ||  $dst eq 'uni' )
+    {
+        my ($un, $cs) = ($sw ? $dst eq 'uni' : $src eq 'uni', $sw ? $src : $dst );
+        $cs = $CP_NAME{$cs} if exists $CP_NAME{$cs};
+        $fn = sprintf $sw ? ($un?'%s2uni':'%s2utf') : ($un?'uni2%s':'utf2%s'), $CODEPAGE{$cs}[0];
+        last if defined &$fn;
+
+        eval "use Unicode::String" unless defined $Unicode::String::VERSION;
+        eval "use Unicode::Map"    unless defined $Unicode::Map::VERSION;
+
+        $CODEPAGE{$cs}[3] = new Unicode::Map( $CODEPAGE{$cs}[1] ) or 
+            die "Can't create Unicode::Map for '$CODEPAGE{$cs}[1]' charset!\n" unless $CODEPAGE{$cs}[3];
+
+        *$fn = eval sprintf $MUTATOR, $un ?
+           sprintf( '%s = $CODEPAGE{%s}[3]->%s_unicode(%s)',
+               $MUTABLE, $cs, (!$sw ? 'from' : 'to'), $MUTABLE ) :
+           sprintf( !$sw ? 
+             '%s = $CODEPAGE{%s}[3]->from_unicode( Unicode::String::utf8(%s) )' :
+             '%s = Unicode::String::utf16(  $CODEPAGE{%s}[3]->to_unicode(%s) )->utf8',
+             $MUTABLE, $cs, $MUTABLE );
+    }
+    else
+    {
+        exists $CP_NAME{$_} and $_ = $CP_NAME{$_} for $src, $dst;
+        exists $CODEPAGE{$_} or die"Unknown codepage '$_'\n" for $src, $dst;
+        $fn = $CODEPAGE{$src}[0].'2'.$CODEPAGE{$dst}[0];
+        last if defined &$fn;
+       *$fn = eval sprintf $MUTATOR, sprintf '(%s) =~ tr/%s/%s/',
+            $MUTABLE, __prepare($src, $dst);
+    }
+
+    return *$fn;
+}
+
+sub __case_factory($$$)
+{
+    my ($cs, $up, $fr, $fn) = @_;
+    $cs = $CP_NAME{$cs} if exists $CP_NAME{$cs};
+    die "Unknown codepage '$cs'\n" if not exists $CODEPAGE{$cs};
+    $fn = ($up?'up':'lo').($fr?'first':'case').'_'.$cs;
+   *$fn = eval sprintf $MUTATOR, sprintf $fr ? 
+        'substr(%s,0,1) =~ tr/%s/%s/' : '(%s) =~ tr/%s/%s/', $MUTABLE, 
+        $up ? unpack('a33a33', $CODEPAGE{$cs}[2]) : reverse unpack('a33a33', $CODEPAGE{$cs}[2])
+        unless defined &$fn;
+    return *$fn;
+}
+
+sub convert($$;$){ 
+    my $fn = __cs2cs_factory shift, shift;
     return &$fn( shift );
 }
 
-sub upcase($;$)
-{
-    my $cs = exists $CP_NAME{$_[0]} ? $CP_NAME{shift()} : shift;
-    die "Unknown codepage '$cs'\n" unless exists $CODEPAGE{$cs}; my $fn = "upcase_$cs";
-    *$fn = eval sprintf $TRANSLATOR, unpack 'a33a33', $CODEPAGE{$cs}[2] unless defined *$fn;
+sub upcase($;$){
+    my $fn = __case_factory shift, 1, 0;
     return &$fn( shift );
 }
 
-sub locase($;$)
-{
-    my $cs = exists $CP_NAME{$_[0]} ? $CP_NAME{shift()} : shift;
-    die "Unknown codepage '$cs'\n" unless exists $CODEPAGE{$cs}; my $fn = "locase_$cs";
-    *$fn = eval sprintf $TRANSLATOR, reverse unpack 'a33a33', $CODEPAGE{$cs}[2] unless defined *$fn;
+sub locase($;$){
+    my $fn = __case_factory shift, 0, 0;
     return &$fn( shift );
 }
 
-sub upfirst($;$)
-{
-    my $cs = exists $CP_NAME{$_[0]} ? $CP_NAME{shift()} : shift;
-    die "Unknown codepage '$cs'\n" unless exists $CODEPAGE{$cs}; my $fn = "upfirst_$cs";
-    *$fn = eval sprintf $TRANSLATOR_FIRST, unpack 'a33a33', $CODEPAGE{$cs}[2] unless defined *$fn;
+sub upfirst($;$){
+    my $fn = __case_factory shift, 1, 1;
     return &$fn( shift );
 }
 
-sub lofirst($;$)
-{
-    my $cs = exists $CP_NAME{$_[0]} ? $CP_NAME{shift()} : shift;
-    die "Unknown codepage '$cs'\n" unless exists $CODEPAGE{$cs}; my $fn = "lofirst_$cs";
-    *$fn = eval sprintf $TRANSLATOR_FIRST, reverse unpack 'a33a33', $CODEPAGE{$cs}[2] unless defined *$fn;
+sub lofirst($;$){
+    my $fn = __case_factory shift, 0, 1;
     return &$fn( shift );
 }
 
-sub charset($)
-{
+sub charset($){
     return $CODEPAGE{shift()}[1];
 }
 
@@ -220,46 +136,38 @@ sub import
     my $self = shift;
 
     if( @_ and exists $CODEPAGE{$_[0]} ){
-        eval q#unless( defined *LC_CTYPE ){ use POSIX 'locale_h' };
-               setlocale LC_CTYPE, 'Russian_Russia.'.shift @_# }
+        eval "use POSIX" unless defined &POSIX::setlocale;
+        POSIX::setlocale( &POSIX::LC_CTYPE, 'Russian_Russia.'.shift @_ );
+    }
 
     return unless @_;
-    my $pkge = caller;
+    my $pkg = caller;
+    no strict qw/refs/;
 
     while( my $src2dst = shift ){
         unless( defined *$src2dst ){
             my ($src, $dst) = $src2dst =~ /^(\w{3})2(\w{3})$/ or
                 die "Unknown import '$src2dst'!\n";
-            exists $CP_NAME{$_} or die"Unknown codepage '$_'\n" for $src, $dst;
-            *$src2dst = eval sprintf $TRANSLATOR, prepare( @CP_NAME{$src, $dst} );
+            __cs2cs_factory( $src, $dst );
         }
-        *{"${pkge}::${src2dst}"} = *$src2dst;
+        *{"${pkg}::${src2dst}"} = *$src2dst;
     }
 }
 
-BEGIN{$TRANSLATOR=<<'END'}
+BEGIN{($MUTABLE, $MUTATOR)=('ref$str?$$str:$str',<<'END')}
 sub(;$){ my $str = $_[0];
-$str = defined wantarray ? $_ : \$_ unless defined $str;
-(ref$str?$$str:$str) =~ tr/%s/%s/;
-return ref $str ? $$str : $str if defined wantarray;
-$_ = $str if defined $_[0] and not ref $str; }
-END
-
-BEGIN{$TRANSLATOR_FIRST=<<'END'}
-sub(;$){ my $str = $_[0];
-$str = defined wantarray ? $_ : \$_ unless defined $str;
-substr(ref$str?$$str:$str,0,1) =~ tr/%s/%s/;
-return ref $str ? $$str : $str if defined wantarray;
+$str = defined wantarray ? $_ : \$_ unless defined $str; %s;
+return ref$str?$$str:$str if defined wantarray;
 $_ = $str if defined $_[0] and not ref $str; }
 END
 
 BEGIN{%CODEPAGE=map{chomp;@_=split/ +/,$_,4;$CP_NAME{$_[1]}=$_[0];$_[0]=>[@_[1..3]]}split/\n/,<<'END'}
-866   dos cp866      񦧨\xffĿٺް´˹cRT
-20866 koi koi8-r     ţ񓛟\xffcRT
-855   ibm cp855      묦窵ޡ쭧諶\xff    Ŀٺ  ´˹                    cRT
-1251  win cp1251     Ũ߯\xa0v--L--L=------+++T++T--TTLL-++
-10007 mac cp10007    ߀\xca                                                  
-28585 iso iso_8859-5 ﰱϤ \xa0                                  cRT
+866   dos ibm866        񦧨\xffĿٺް´˹cRT
+20866 koi koi8-r        ţ񓛟\xffcRT
+855   ibm cp855         묦窵ޡ쭧諶\xff    Ŀٺ  ´˹                    cRT
+1251  win windows-1251  Ũ߯\xa0v--L--L=------+++T++T--TTLL-++
+10007 mac ms-cyrillic   ߀\xca                                                  
+28585 iso iso_8859-5    ﰱϤ \xa0                                  cRT
 END
 
 BEGIN{$STATISTIC=<<'END'}
@@ -300,3 +208,195 @@ BEGIN{$STATISTIC=<<'END'}
 END
 
 1;
+
+__END__
+
+=head1 NAME
+
+cyrillic - Library for fast and easy cyrillic text manipulation
+
+=head1 SYNOPSIS
+
+  use cyrillic qw/866 win2dos convert locase upcase detect/;
+
+  print convert( 866, 1251, $str );
+  print convert( 'dos','win', \$str );
+  print win2dos $str;
+
+=head1 DESCRIPTION
+
+This module includes cyrillic string converting functions 
+from one and to another charset, to upper and to lower case without
+locale switching. Also included single-byte charsets detection routine.
+It is easy to add new code pages. For this purpose it is necessary 
+only to add appropriate string of a code page.
+
+Supported charsets:
+    ibm866, koi8-r, cp855, windows-1251, MacWindows, iso_8859-5, unicode, utf8;
+
+If the first imported parameter - number of a code page, then locale will be switched to it.
+
+NOTE! Specialisation (like B<win2dos>, B<utf2win>) call faster then B<convert>.
+
+NOTE! Only B<convert> function and they specialisation work with Unicode and UTF-8 strings.
+All others function work only with single-byte sharsets.
+
+Names for using in named charset convertors:
+
+    dos ibm866       866
+    koi koi8-r       20866
+    ibm cp855        855
+    win windows-1251 1251 
+    mac ms-cyrillic  10007
+    iso iso_8859-5   28585
+    uni Unicode
+    utf UTF-8
+
+
+=head1 FUNCTIONS
+
+Library includes converting and helper functions:
+    convert, upcase, locase, upfirst, lofirst;
+    detect, charset.
+
+
+At importing list might be listed named convertors. For Ex.:
+
+  use cyrillic qw/dos2win win2koi mac2dos ibm2dos/;
+
+
+The following rules are correct for converting functions:
+
+  VAR may be SCALAR or REF to SCALAR.
+  If VAR is REF to SCALAR then SCALAR will be converted.
+  If VAR is ommited then $_ operated.
+  If function called to void context and VAR is not REF
+  then result placed to $_.
+
+
+=head1 CONVERSION METHODS
+
+=item B<convert> SRC_CP, DST_CP, [VAR]
+
+Convert VAR from SRC_CP codepage to DST_CP codepage and returns
+converted string.
+  
+The converting Unicode or UTF-8 data requires presence of 
+installed Unicode::String and Unicode::Map.
+
+=item B<upcase> CODEPAGE, [VAR]
+
+Convert VAR to uppercase using CODEPAGE table and returns
+converted string.
+
+=item B<locase> CODEPAGE, [VAR]
+
+Convert VAR to lowercase using CODEPAGE table and returns
+converted string.
+
+=item B<upfirst> CODEPAGE, [VAR]
+
+Convert first char of VAR to uppercase using CODEPAGE table and returns
+converted string.
+
+=item B<lofirst> CODEPAGE, [VAR]
+
+Convert first char of VAR to lowercase using CODEPAGE table and returns
+converted string.
+
+=head1 MAINTAINANCE METHODS
+
+=item B<charset> CODEPAGE
+
+Returns charset name for CODEPAGE.
+
+=item B<detect> ARRAY
+
+Detect single-byte codepage of data in ARRAY and returns codepage number.
+If codepage not detected then returns undefined value;
+
+=head1 EXAMPLES
+
+  use cyrillic qw/convert locase upcase detect dos2win win2dos/;
+
+  $\ = "\n";
+  $_ = "\x8F\xE0\xA8\xA2\xA5\xE2 \xF0\xA6\x88\xAA\x88!";
+
+  print; upcase 866;
+  print; dos2win;
+  print; win2dos;
+  print; locase 866;
+  print;
+  print detect $_;
+
+
+  
+  # CONVERTING TEST:
+
+  use cyrillic qw/utf2dos mac2utf dos2mac win2dos utf2win/;
+
+  $_ = "Хелло Ворльд!\n";
+
+  print "UTF-8: $_";
+  print "  DOS: ", utf2dos mac2utf dos2mac win2dos utf2win $_;
+
+
+
+  # EQVIVALENT CALLS:
+
+  dos2win( $str );        # called to void context -> result placed to $_
+  $_ = dos2win( $str );
+
+  dos2win( \$str );       # called with REF to string -> direct converting
+  $str = dos2win( $str );
+
+  dos2win();              # with ommited param called -> $_ converted
+  dos2win( \$_ );
+  $_ = dos2win( $_ );
+
+
+  
+  # FOR EASY SWITCH LOCALE CODEPAGE
+
+  use cyrillic qw/866/;   # locale switched to Russian_Russia.866
+
+  use locale;
+  print $str =~ /(\w+)/;
+
+  no locale;
+  print $str =~ /(\w+)/;
+
+=head1 FAQ
+
+  * Q: Why module say: Can't create Unicode::Map for 'koi8-r' charset!
+    A: Your Unicode::Map module can't find map file for 'koi8-r' charset.
+       In Unicode::Map manual is told whence it is possible to download 
+       this file and as it to install in the system.
+
+  * Q: Why perl say: "Undefined subroutine koi2win called" ?
+    A: The function B<koi2win> is specialization of the function B<convert>, 
+       which is created at inclusion it of the name in the list of import.
+
+
+=head1 AUTHOR
+
+Albert MICHEEV <Albert@f80.n5049.z2.fidonet.org>
+
+=head1 COPYRIGHT
+
+Copyright (C) 2000, Albert MICHEEV
+
+This module is free software; you can redistribute it or modify it
+under the same terms as Perl itself.
+
+=head1 AVAILABILITY
+
+The latest version of this library is likely to be available from:
+
+http://www.perl.com/CPAN
+
+=head1 SEE ALSO
+
+Unicode::String, Unicode::Map.
+
+=cut
